/*
   Copyright (C) 2000-2002  Ulric Eriksson <ulric@siag.nu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
*/

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>

#include "config.h"

struct query {
	char *name, *value;
	struct query *next;
} *queries = NULL;

static int cgi = 0;
static struct query *address, *port;

static void error(char *fmt, ...)
{
	char b[4096];
	va_list ap;
	va_start(ap, fmt);
	vsnprintf(b, sizeof b, fmt, ap);
	if (cgi) printf("%s\n", b);
	else fprintf(stderr, "%s\n", b);
	va_end(ap);
	if (cgi) exit(0);
	else exit(1);
}

static void alarm_handler(int dummy)
{
	;
}

static void usage(void)
{
	printf("usage: penctl host:port command\n");
	exit(0);
}

static void parse_query(char *q)
{
	char *p, *r;
	struct query *m;
	q = strdup(q);
	for (p = strtok(q, "&"); p; p = strtok(NULL, "&")) {
		m = malloc(sizeof *m);
		m->name = p;
		r = strchr(p, '=');
		if (r) m->value = r+1;
		else m->value = NULL;
		m->next = queries;
		queries = m;
	}
}

static struct query *get_query(char *name)
{
	struct query *m;
	for (m = queries; m; m = m->next) {
		if (!strcmp(name, m->name)) return m;
	}
	return NULL;
}

static struct query *require_query(char *name)
{
	struct query *m = get_query(name);
	if (m == NULL) error("%s: no such query");
	return m;
}

static void header(char *title)
{
	printf("<html><head><title>Penctl: %s</title></head>\n", title);
	printf("<body><h1>%s</h1>\n", title);
}

static void footer(void)
{
	printf("</body></html>\n");
}

static void page_main(void)
{
	header("Main");
	printf("<form><table><td>\n");
	printf("<tr><td>Address<td><input type=\"text\" name=\"address\">\n");
	printf("<tr><td>Port<td><input type=\"text\" port=\"port\">\n");
	printf("</table>\n");
	printf("<input type=\"submit\"></form>\n");
	footer();
}

static void page_servers(void)
{
	header("Servers");
	footer();
}

static void page_clients(void)
{
	header("Clients");
	footer();
}

static void page_mode(void)
{
	header("Mode");
	footer();
}

static void page_status(void)
{
	header("Status");
	footer();
}

static int open_socket(char *addr, char *port)
{
	int fd = -1;
	struct sockaddr_in serv_addr;
	struct hostent *h;
	struct in_addr a;
	struct servent *s;
	int n, po;
	h = gethostbyname(addr);
	if (h == NULL) error("unknown or invalid address %s\n", addr);
	memcpy(&a, h->h_addr, h->h_length);
	s = getservbyname(port, "tcp");
	if (s == NULL) po = atoi(port);
	else po = ntohs(s->s_port);

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0) error("error opening socket");
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = a.s_addr;
	serv_addr.sin_port = htons(po);
	signal(SIGALRM, alarm_handler);
	n = connect(fd, (struct sockaddr *)&serv_addr, sizeof serv_addr);
	alarm(0);
	if (n == -1) {
		error("error connecting to server");
	}
	return fd;
}

static void do_cgi(int fd)
{
	struct query *mode = get_query("mode");
	if (mode == NULL || !strcmp(mode->value, "main")) page_main();
	else if (!strcmp(mode->value, "servers")) page_servers();
	else if (!strcmp(mode->value, "clients")) page_clients();
	else if (!strcmp(mode->value, "mode")) page_mode();
	else if (!strcmp(mode->value, "status")) page_status();
	else page_main();
}

static void dont_cgi(int fd, char **argv)
{
	int i, n;
	char b[1024];
	strcpy(b, argv[2]);
	for (i = 3; argv[i]; i++) {
		strcat(b, " ");
		strcat(b, argv[i]);
	}
	strcat(b, "\n");
	write(fd, b, strlen(b));
	for (;;) {
		n = read(fd, b, sizeof b);
		if (n == 0) break;
		if (n == -1) error("error reading");
		write(1, b, n);
	}
	close(fd);
}

int main(int argc, char **argv)
{
	char *addr, *p;
	char b[1024];

	cgi = (getenv("GATEWAY_INTERFACE") != NULL);

	if (cgi) {
		char *q = getenv("QUERY_STRING");
		if (q) parse_query(getenv("QUERY_STRING"));
		printf("Content-type: text/html\r\n\r\n");
		fflush(stdout);
	} else if (argc < 3) {
		usage();
	}

	if (cgi) {
		address = get_query("address");
		port = get_query("port");
		if (address == NULL || port == NULL) {
			page_main();
			return 0;
		}
		do_cgi(open_socket(address->value, port->value));
	} else {
		strcpy(b, argv[1]);
		addr = b;
		p = strchr(b, ':');
		if (p == NULL) error("no port given");

		*p++ = '\0';

		dont_cgi(open_socket(addr, p), argv);
	}

	return 0;
}
